import cyvlfeat.fisher
import cyvlfeat.generic
import cyvlfeat.gmm
import cyvlfeat.hog
import cyvlfeat.kmeans
import cyvlfeat.sift


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
